from yaml.loader import SafeLoader

from mazikeen.GeneratorException import GeneratorException


class SafeLineLoader(SafeLoader):
    def construct_mapping(self, node, deep=False):
        mapping = super(SafeLineLoader, self).construct_mapping(node, deep=deep)
        # Add 1 so line numbering starts at 1
        mapping['__line__'] = node.start_mark.line + 1
        return mapping
        

def getYamlInt(data, line, field):
    if (not isinstance(data, int)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects an integer at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects an integer at line {line}")

    return data

def getYamlIntOrNone(data, line, field):
    if isinstance(data, str):
        if data.lower() == "none": return None
    if (not isinstance(data, int)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects an integer at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects an integer at line {line}")

    return data

def getYamlBool(data, line, field):
    if (not isinstance(data, bool)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects a bool at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects a bool at line {line}")
    return data

def getYamlString(data, line, field):
    if (not isinstance(data, str)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects an integer at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects a string at line {line}")
    return data

def getYamlList(data, line, field):
    if (not isinstance(data, list)):
        if (isinstance(data, dict)):
            raise GeneratorException(f"field '{field}' expects a list at line {data['__line__']}")
        raise GeneratorException(f"field '{field}' expects a list at line {line}")
    return data


import pathlib

from mazikeen.Utils import replaceVariables, ensure_dir
from mazikeen.ConsolePrinter import Printer

class MakedirsBlock:
    def __init__(self, dir):
        self.dir = dir

    def run(self, workingDir = "", variables = {}, printer = Printer()):
        printer.verbose("Makedirs:", self.dir)
        _dir = replaceVariables(self.dir, variables, printer)
        _dir = str(pathlib.PurePath(workingDir).joinpath(_dir)) + "/" 
        ensure_dir(_dir)
        return True


import io
import threading

class Printer():
    printerCnt = 0
    printLock = threading.Lock()
    errorLock = threading.Lock()
    def __init__(self, verbose = False):
        self._verbose = verbose
        self.__printerCnt = Printer.printerCnt
        Printer.printerCnt += 1
        self.errorOutput = ""

    def __repr__(self):
        return f"{self.__class__.__name__} {self.__printerCnt}"

    def __str__(self):
        return self.__repr__()

    def verbose(self, *args, **kwargs):
        if self._verbose:
            self.print(*args, **kwargs, flush = True)
    
    def getBufferedPrinter(self):
        bp = BufferedPrinter(verbose = self._verbose)
        bp.errorOutput = self.errorOutput
        return bp
    
    def print(self, *args, **kwargs):
        with Printer.printLock:
            print(*args, **kwargs)
    
    def debug(self, *args, **kwargs):
        self.print("debug: ", end ="", flush = False) 
        self.print(*args, **kwargs, flush = True)

    def error(self, *args, **kwargs):
        with Printer.errorLock:
            with io.StringIO() as _errorMsg:
                print("Error: ", file=_errorMsg, end = "", flush = False)
                print(*args, **kwargs, file=_errorMsg, flush = True)
                self.errorOutput +=_errorMsg.getvalue()
            self.print("Error: ", end = "", flush = False)
            self.print(*args, **kwargs, flush = True)
        
    def flush(self):
        return
        
    def isVerbose(self):
        return self._verbose
        
class BufferedPrinter(Printer):
    def __init__(self, verbose = False):
        super().__init__(verbose)
        self.__buffer = None
        
    def print(self, *args, **kwargs):
        output = io.StringIO()
        print(*args, file=output, **kwargs)
        outputStr = output.getvalue()
        if not self.__buffer:
            self.__buffer = outputStr
        else :
            self.__buffer += outputStr
        
    def verbose(self, *args, **kwargs):
        if self._verbose:
            self.print(*args, **kwargs, flush = True)

    def flush(self):
        if self.__buffer:
            super().print(self.__buffer, end="", flush = True)
        self.__buffer = None
        
    def getBufferedPrinter(self):
        return self
    
    def __del__(self):
        self.flush()



import shlex
import subprocess
import os
import pathlib

from mazikeen.ConsolePrinter import Printer
from mazikeen.Utils import replaceVariables, ensure_dir

class RunBlock:
    def __init__(self, cmd, outputfile = None, inputfile = None, exitcode = None):
        self.cmd = cmd
        self.outputfile = outputfile
        self.inputfile = inputfile
        self.exitcode = exitcode

    def run(self, workingDir = ".", variables = {}, printer = Printer()):
        printer.verbose("Run:", self.cmd)
        replCmd = replaceVariables(self.cmd, variables)
        cmdNArgs = shlex.split(replCmd)
        printer.verbose("cwd:", os.getcwd())
        printer.verbose("call:", replCmd)
        inputfileData = None
        if self.inputfile:
            with open(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.inputfile, variables)), "rb") as fh:
                inputfileData = fh.read()
        subProcessRes = subprocess.run(cmdNArgs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, input=inputfileData, cwd = workingDir, shell = False)
        if self.outputfile:
            outputfileFullPath = str(pathlib.PurePath(workingDir).joinpath(replaceVariables(self.outputfile, variables)))
            ensure_dir(outputfileFullPath)
            with open(outputfileFullPath, "wb") as fh:
                fh.write(subProcessRes.stdout)
        
        res = True
        if (self.exitcode != None):
            res = subProcessRes.returncode == self.exitcode
            if not res:
                printer.error("different exitcode received:", subProcessRes.returncode, "!=", self.exitcode, "for command '"+ str(replCmd) +"'")
        return res
     


import yaml
from yaml.loader import SafeLoader

from mazikeen.GeneratorLooper import serialReadYaml

class SafeLineLoader(SafeLoader):
    def construct_mapping(self, node, deep=False):
        mapping = super(SafeLineLoader, self).construct_mapping(node, deep=deep)
        # Add 1 so line numbering starts at 1
        mapping['__line__'] = node.start_mark.line + 1
        return mapping

with open('script_test.yaml') as f:
    data = yaml.load(f, Loader=SafeLineLoader)
    print ("data = ", data)
    res = serialReadYaml(data)
    pass


import re
import os
import shutil
import stat
import pathlib
from enum import Enum

from mazikeen.ConsolePrinter import Printer
from mazikeen.GeneratorException import GeneratorException

__replaceVariablesExp = re.compile(r'(?<!\\)\${.*?}')
def replaceVariables(line, dictReplVar, printer = Printer()):
    global __replaceVariablesExp
    if line == None: return (True, line)
    searchStart = 0
    while (True):
        m = __replaceVariablesExp.search(line[searchStart:])
        if not m: 
            break
        foundVar = m.group()[2:-1]
        replaceVal = dictReplVar.get(foundVar)
        if (replaceVal):
            strReplaceVal = str(replaceVal)
            line = line[:searchStart + m.start()] + strReplaceVal + line[searchStart + m.end():]
            searchStart += m.start() + len(strReplaceVal)
        else:
            raise GeneratorException("Variable " + str(m.group()) + " does not exist")
    line = line.replace(r'\$', "$")
    return line

def ensure_dir(file_path):
    directory = os.path.dirname(file_path)
    if (directory) and (not os.path.exists(directory)):
        os.makedirs(directory)

def rmtree(path, printer = Printer()):
    try:
        def remove_readonly(fn, path, excinfo):
            os.chmod(path, stat.S_IWRITE)
            fn(path)
        if os.path.exists(path):
            shutil.rmtree(path, onerror=remove_readonly, ignore_errors = False)
        return True
    except Exception as e:
        printer.error("rmtree:", e)
        return False

class diffStrategy(Enum):
   IgnoreLeftOrphans = 0
   IgnoreRightOrphans = 1
   IgnoreOrphans = 2
   All = 3

def __listAllFilesWithoutRoot(path):
    allFilesWithRoot = [pathlib.Path(dp) / f for dp, dn, fn in os.walk(os.path.expanduser(path)) for f in fn + dn]
    rootMathLen = len(pathlib.Path(path).parts)
    filesWithoutRoot = [pathlib.Path(*f.parts[rootMathLen:]) for f in allFilesWithRoot]
    return filesWithoutRoot
    
def __getCompareLine(line, fh, compiledIgnoreLines):
    repeate = True
    while(repeate):
        if not line: return line
        repeate = False
        try:
            strLine = line.decode('utf-8')
        except:
            return line
        for ignoreLine in compiledIgnoreLines:
            if ignoreLine.match(strLine):
                line = fh.readline()
                repeate = True
                break
    return line

def normalizeEOL(line):
    # Todo: can be optimized
    if len(line) >=2 and line[-2:] == bytes('\r\n', 'utf-8'): 
        line = line[:-2] + bytes('\n', 'utf-8')
    elif len(line) >=1 and line[-1] == bytes('\r', 'utf-8'): 
        line = line[:-1] + bytes('\n', 'utf-8')
    return line

def diffFiles(fileL, fileR, compiledIgnoreLines = [], binaryCompare = False):
    def areLinesIdentical(lineL, lineR, binaryCompare):
        def normalizeEOL(line):
            if len(line) >=2 and line[-2:] == bytes('\r\n', 'utf-8'): 
                line = line[:-2] + bytes('\n', 'utf-8')
            elif len(line) >=1 and line[-1] == bytes('\r', 'utf-8'): 
                line = line[:-1] + bytes('\n', 'utf-8')
            return line

        if lineL == lineR: return True #This is the most likely scenario so it has priority
        if binaryCompare == False:
            return normalizeEOL(lineL) == normalizeEOL(lineR)
        return False

    with open(fileL, "rb") as fhL:
        with open(fileR, "rb") as fhR:
            while True:
                lineL = fhL.readline()
                lineR = fhR.readline()
                if (not lineL and not lineR): break #EOF reached
                if areLinesIdentical(lineL, lineR, binaryCompare) == False:
                    lineL = __getCompareLine(lineL, fhL, compiledIgnoreLines)
                    lineR = __getCompareLine(lineR, fhR, compiledIgnoreLines)
                    if areLinesIdentical(lineL, lineR, binaryCompare) == False:
                        return False
    return True

def diff(pathL, pathR, binaryCompare = False, diffStrategy = diffStrategy.All, ignoreLines = [], printer = Printer(verbose = True)):
    rootM = pathlib.Path(pathL)
    rootS = pathlib.Path(pathR)
    compiledIgnoreLines = list(map(lambda x: re.compile(x), ignoreLines))
    if (rootM.is_file() and rootS.is_file()):
        if (not diffFiles(rootM, rootS, compiledIgnoreLines = compiledIgnoreLines, binaryCompare = binaryCompare)):
            printer.error(f"diff failed: '{rootM}' != '{rootS}'")
            return False
        return True

    for file in [rootM, rootS]:
        if not file.exists():
            printer.error(f"diff failed: '{file}' doesn't exist")
            return False

    filesM = set(__listAllFilesWithoutRoot(pathL))
    filesS = set(__listAllFilesWithoutRoot(pathR))
    
    if diffStrategy == diffStrategy.IgnoreLeftOrphans:
        filesM, filesS = filesS, filesS
        rootM, rootS = rootS, rootM
    
    if diffStrategy != diffStrategy.IgnoreOrphans:
        for file in (filesM - filesS):
            printer.error(f"diff failed: '{str(rootM/file)}' not in '{rootS / file.parent}'")
            return False
    
    if diffStrategy == diffStrategy.All:
        for file in (filesS - filesM):
            printer.error(f"diff failed: '{str(rootS/file)}' not in '{rootM / file.parent}'")
            return False
    
    comPaths =[(rootM/path, rootS/path) for path in (filesM & filesS)]
    
    for pathM, pathS in comPaths:
        if (pathM.is_file() != pathS.is_file()):
            printer.error(f"diff failed: '{pathM}' != '{pathS}'")
            return False
    

    comFiles = [(pathM, pathS) for pathM, pathS in comPaths if pathM.is_file()]
    for fileM, fileS in comFiles:
        if (not diffFiles(fileM, fileS, binaryCompare = binaryCompare, compiledIgnoreLines = compiledIgnoreLines)):
            printer.error(f"diff failed: '{fileM}' != '{fileS}'")
            return False
    return True




from mazikeen.RmdirBlock import RmdirBlock
from mazikeen.GeneratorException import GeneratorException
from mazikeen.ConsolePrinter import Printer, BufferedPrinter

def generateRmdir(data):
    if not isinstance(data, str):
        raise GeneratorException(f"'rmdirs' block not recognized")
    return RmdirBlock(data)


import signal

__signal_received__ = None

def failFast():
    global __signal_received__
    return __signal_received__ != None

def init(): 
    signal.signal(signal.SIGINT, signalHandler)

def signalName():
    global __signal_received__
    if __signal_received__ == None: return __signal_received__
    return str(signal.Signals(__signal_received__).name)

def signalHandler(signal_received, frame):
    global __signal_received__
    if (__signal_received__ == None):
        print(str(signal.Signals(signal_received).name) + " received fail fast enabled. Send signal again to force exit")
        __signal_received__ = signal_received
    else:
        exit(1)
    